@@ -1,4 +1,4 @@
-/* Copyright (C) 2015-2016 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
+/* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
 
 #include "routingtable.h"
 #include "peer.h"
@@ -17,8 +17,10 @@ static inline u8 bit_at(const u8 *key, u8 a, u8 b)
 {
 	return (key[a] >> b) & 1;
 }
-static inline void assign_cidr(struct routing_table_node *node, u8 cidr)
+static inline void copy_and_assign_cidr(struct routing_table_node *node, const u8 *src, u8 cidr)
 {
+	memcpy(node->bits, src, (cidr + 7) / 8);
+	node->bits[(cidr + 7) / 8 - 1] &= 0xff << ((8 - (cidr % 8)) % 8);
 	node->cidr = cidr;
 	node->bit_at_a = cidr / 8;
 	node->bit_at_b = 7 - (cidr % 8);
@@ -32,9 +34,13 @@ static inline void assign_cidr(struct routing_table_node *node, u8 cidr)
  *       return;
  *     free_node(node->bit[0]);
  *     free_node(node->bit[1]);
- *     kfree_rcu(node);
+ *     kfree_rcu_bh(node);
  * }
  */
+static void node_free_rcu(struct rcu_head *rcu)
+{
+	kfree(container_of(rcu, struct routing_table_node, rcu));
+}
 #define ref(p) rcu_access_pointer(p)
 #define push(p) do { BUG_ON(len >= 128); stack[len++] = rcu_dereference_protected(p, lockdep_is_held(lock)); } while (0)
 static void free_node(struct routing_table_node *top, struct mutex *lock)
@@ -59,7 +65,7 @@ static void free_node(struct routing_table_node *top, struct mutex *lock)
 			if (ref(node->bit[1]))
 				push(node->bit[1]);
 		} else {
-			kfree_rcu(node, rcu);
+			call_rcu_bh(&node->rcu, node_free_rcu);
 			--len;
 		}
 		prev = node;
@@ -183,7 +189,7 @@ static inline struct routing_table_node *find_node(struct routing_table_node *tr
 			found = node;
 		if (node->cidr == bits)
 			break;
-		node = rcu_dereference(node->bit[bit_at(key, node->bit_at_a, node->bit_at_b)]);
+		node = rcu_dereference_bh(node->bit[bit_at(key, node->bit_at_a, node->bit_at_b)]);
 	}
 	return found;
 }
@@ -208,17 +214,13 @@ static inline bool node_placement(struct routing_table_node __rcu *trie, const u
 static int add(struct routing_table_node __rcu **trie, u8 bits, const u8 *key, u8 cidr, struct wireguard_peer *peer, struct mutex *lock)
 {
 	struct routing_table_node *node, *parent, *down, *newnode;
-	int bits_in_common;
 
 	if (!rcu_access_pointer(*trie)) {
 		node = kzalloc(sizeof(*node) + (bits + 7) / 8, GFP_KERNEL);
 		if (!node)
 			return -ENOMEM;
 		node->peer = peer;
-		memcpy(node->bits, key, (cidr + 7) / 8);
-		/* Not strictly neccessary for the data structure, but helps keep the data cleaner: */
-		node->bits[(cidr + 7) / 8 - 1] &= 0xff << ((8 - (cidr % 8)) % 8);
-		assign_cidr(node, cidr);
+		copy_and_assign_cidr(node, key, cidr);
 		rcu_assign_pointer(*trie, node);
 		return 0;
 	}
@@ -233,10 +235,7 @@ static int add(struct routing_table_node __rcu **trie, u8 bits, const u8 *key, u
 	if (!newnode)
 		return -ENOMEM;
 	newnode->peer = peer;
-	memcpy(newnode->bits, key, (cidr + 7) / 8);
-	/* Not strictly neccessary for the data structure, but helps keep the data cleaner: */
-	newnode->bits[(cidr + 7) / 8 - 1] &= 0xff << ((8 - (cidr % 8)) % 8);
-	assign_cidr(newnode, cidr);
+	copy_and_assign_cidr(newnode, key, cidr);
 
 	if (!node)
 		down = rcu_dereference_protected(*trie, lockdep_is_held(lock));
@@ -247,15 +246,13 @@ static int add(struct routing_table_node __rcu **trie, u8 bits, const u8 *key, u
 		return 0;
 	}
 	/* here we must be inserting between node and down */
-	bits_in_common = common_bits(down, key, cidr);
+	cidr = min(cidr, common_bits(down, key, cidr));
 	parent = node;
-	if (bits_in_common > cidr)
-		bits_in_common = cidr;
 
 	/* we either need to make a new branch above down and newnode
 	 * or newnode can be the branch. newnode can be the branch if
 	 * its cidr == bits_in_common */
-	if (newnode->cidr == bits_in_common) {
+	if (newnode->cidr == cidr) {
 		/* newnode can be the branch */
 		rcu_assign_pointer(newnode->bit[bit_at(down->bits, newnode->bit_at_a, newnode->bit_at_b)], down);
 		if (!parent)
@@ -269,9 +266,9 @@ static int add(struct routing_table_node __rcu **trie, u8 bits, const u8 *key, u
 			kfree(newnode);
 			return -ENOMEM;
 		}
-		assign_cidr(node, bits_in_common);
 		node->incidental = true;
-		memcpy(node->bits, newnode->bits, (bits + 7) / 8);
+		copy_and_assign_cidr(node, newnode->bits, cidr);
+
 		rcu_assign_pointer(node->bit[bit_at(down->bits, node->bit_at_a, node->bit_at_b)], down);
 		rcu_assign_pointer(node->bit[bit_at(newnode->bits, node->bit_at_a, node->bit_at_b)], newnode);
 		if (!parent)
@@ -283,7 +280,7 @@ static int add(struct routing_table_node __rcu **trie, u8 bits, const u8 *key, u
 }
 
 #define push(p) do { \
-	struct routing_table_node *next = (maybe_lock ? rcu_dereference_protected(p, lockdep_is_held(maybe_lock)) : rcu_dereference(p)); \
+	struct routing_table_node *next = (maybe_lock ? rcu_dereference_protected(p, lockdep_is_held(maybe_lock)) : rcu_dereference_bh(p)); \
 	if (next) { \
 		BUG_ON(len >= 128); \
 		stack[len++] = next; \
@@ -392,11 +389,11 @@ inline struct wireguard_peer *routing_table_lookup_v4(struct routing_table *tabl
 	struct wireguard_peer *peer = NULL;
 	struct routing_table_node *node;
 
-	rcu_read_lock();
-	node = find_node(rcu_dereference(table->root4), 32, (const u8 *)ip);
+	rcu_read_lock_bh();
+	node = find_node(rcu_dereference_bh(table->root4), 32, (const u8 *)ip);
 	if (node)
 		peer = peer_get(node->peer);
-	rcu_read_unlock();
+	rcu_read_unlock_bh();
 	return peer;
 }
 
@@ -406,11 +403,11 @@ inline struct wireguard_peer *routing_table_lookup_v6(struct routing_table *tabl
 	struct wireguard_peer *peer = NULL;
 	struct routing_table_node *node;
 
-	rcu_read_lock();
-	node = find_node(rcu_dereference(table->root6), 128, (const u8 *)ip);
+	rcu_read_lock_bh();
+	node = find_node(rcu_dereference_bh(table->root6), 128, (const u8 *)ip);
 	if (node)
 		peer = peer_get(node->peer);
-	rcu_read_unlock();
+	rcu_read_unlock_bh();
 	return peer;
 }
 
@@ -446,28 +443,28 @@ int routing_table_remove_by_peer(struct routing_table *table, struct wireguard_p
 int routing_table_walk_ips(struct routing_table *table, void *ctx, int (*func)(void *ctx, struct wireguard_peer *peer, union nf_inet_addr ip, u8 cidr, int family))
 {
 	int ret;
-	rcu_read_lock();
-	ret = walk_ips(rcu_dereference(table->root4), AF_INET, ctx, func, NULL);
-	rcu_read_unlock();
+	rcu_read_lock_bh();
+	ret = walk_ips(rcu_dereference_bh(table->root4), AF_INET, ctx, func, NULL);
+	rcu_read_unlock_bh();
 	if (ret)
 		return ret;
-	rcu_read_lock();
-	ret = walk_ips(rcu_dereference(table->root6), AF_INET6, ctx, func, NULL);
-	rcu_read_unlock();
+	rcu_read_lock_bh();
+	ret = walk_ips(rcu_dereference_bh(table->root6), AF_INET6, ctx, func, NULL);
+	rcu_read_unlock_bh();
 	return ret;
 }
 
 int routing_table_walk_ips_by_peer(struct routing_table *table, void *ctx, struct wireguard_peer *peer, int (*func)(void *ctx, union nf_inet_addr ip, u8 cidr, int family))
 {
 	int ret;
-	rcu_read_lock();
-	ret = walk_ips_by_peer(rcu_dereference(table->root4), AF_INET, ctx, peer, func, NULL);
-	rcu_read_unlock();
+	rcu_read_lock_bh();
+	ret = walk_ips_by_peer(rcu_dereference_bh(table->root4), AF_INET, ctx, peer, func, NULL);
+	rcu_read_unlock_bh();
 	if (ret)
 		return ret;
-	rcu_read_lock();
-	ret = walk_ips_by_peer(rcu_dereference(table->root6), AF_INET6, ctx, peer, func, NULL);
-	rcu_read_unlock();
+	rcu_read_lock_bh();
+	ret = walk_ips_by_peer(rcu_dereference_bh(table->root6), AF_INET6, ctx, peer, func, NULL);
+	rcu_read_unlock_bh();
 	return ret;
 }
 
