@@ -1,4 +1,4 @@
-/* Copyright (C) 2015-2016 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
+/* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
 
 #ifdef __linux__
 #include <libmnl/libmnl.h>
@@ -18,7 +18,6 @@
 #include <unistd.h>
 #include <time.h>
 #include <dirent.h>
-#include <poll.h>
 #include <signal.h>
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -41,7 +40,16 @@ struct inflatable_buffer {
 	size_t pos;
 };
 
-#define max(a, b) (a > b ? a : b)
+#define max(a, b) ((a) > (b) ? (a) : (b))
+
+static int check_version_magic(struct wgdevice *device, int ret)
+{
+	if (ret == -EPROTO || (!ret && device->version_magic != WG_API_VERSION_MAGIC)) {
+		fprintf(stderr, "This program was built for a different version of WireGuard than\nwhat is currently running. Either this version of wg(8) is out\nof date, or the currently loaded WireGuard module is out of date.\nIf you have just updated your WireGuard installation, you may have\nforgotten to unload the previous running WireGuard module. Try\nrunning `rmmod wireguard` as root, and then try re-adding the device\nand trying again.\n\n");
+		return -EPROTO;
+	}
+	return ret;
+}
 
 static int add_next_to_inflatable_buffer(struct inflatable_buffer *buffer)
 {
@@ -173,12 +181,17 @@ static int userspace_set_device(struct wgdevice *dev)
 	ret = -EBADMSG;
 	if (!len)
 		goto out;
+	dev->version_magic = WG_API_VERSION_MAGIC;
 	ret = write(fd, dev, len);
 	if (ret < 0)
 		goto out;
 	ret = read(fd, &ret_code, sizeof(ret_code));
 	if (ret < 0)
 		goto out;
+	if (ret != sizeof(ret_code)) {
+		ret = -EBADMSG;
+		goto out;
+	}
 	ret = ret_code;
 out:
 	close(fd);
@@ -186,68 +199,78 @@ out:
 	return (int)ret;
 }
 
+#define READ_BYTES(bytes) ({ \
+	void *__p; \
+	size_t __bytes = (bytes); \
+	if (bytes_left < __bytes) { \
+		offset = p - buffer; \
+		bytes_left += buffer_size; \
+		buffer_size *= 2; \
+		ret = -ENOMEM; \
+		p = realloc(buffer, buffer_size); \
+		if (!p) \
+			goto out; \
+		buffer = p; \
+		p += offset; \
+	} \
+	bytes_left -= __bytes; \
+	ret = read(fd, p, __bytes); \
+	if (ret < 0) \
+		goto out; \
+	if ((size_t)ret != __bytes) { \
+		ret = -EBADMSG; \
+		goto out; \
+	} \
+	__p = p; \
+	p += __bytes; \
+	__p; \
+})
 static int userspace_get_device(struct wgdevice **dev, const char *interface)
 {
-	struct pollfd pollfd = { .events = POLLIN };
-	int len;
-	char byte = 0;
-	size_t i;
-	struct wgpeer *peer;
+	unsigned int len = 0, i;
+	size_t buffer_size, bytes_left;
 	ssize_t ret;
+	ptrdiff_t offset;
+	uint8_t *buffer = NULL, *p, byte = 0;
+
 	int fd = userspace_interface_fd(interface);
 	if (fd < 0)
 		return fd;
-	*dev = NULL;
+
 	ret = write(fd, &byte, sizeof(byte));
 	if (ret < 0)
 		goto out;
-
-	pollfd.fd = fd;
-	if (poll(&pollfd, 1, -1) < 0)
-		goto out;
-	ret = -ECONNABORTED;
-	if (!(pollfd.revents & POLLIN))
-		goto out;
-
-	ret = ioctl(fd, FIONREAD, &len);
-	if (ret < 0) {
-		ret = -errno;
-		goto out;
-	}
-	ret = -EBADMSG;
-	if ((size_t)len < sizeof(struct wgdevice))
-		goto out;
-
-	ret = -ENOMEM;
-	*dev = malloc(len);
-	if (!*dev)
-		goto out;
-
-	ret = read(fd, *dev, len);
-	if (ret < 0)
-		goto out;
-	if (ret != len) {
+	if (ret != sizeof(byte)) {
 		ret = -EBADMSG;
 		goto out;
 	}
 
-	ret = -EBADMSG;
-	for_each_wgpeer(*dev, peer, i) {
-		if ((uint8_t *)peer + sizeof(struct wgpeer) > (uint8_t *)*dev + len)
-			goto out;
-		if ((uint8_t *)peer + sizeof(struct wgpeer) + sizeof(struct wgipmask) * peer->num_ipmasks > (uint8_t *)*dev + len)
+	ioctl(fd, FIONREAD, &len);
+	bytes_left = buffer_size = max(len, sizeof(struct wgdevice) + sizeof(struct wgpeer) + sizeof(struct wgipmask));
+	p = buffer = malloc(buffer_size);
+	ret = -ENOMEM;
+	if (!buffer)
 		goto out;
-	}
+
+	len = ((struct wgdevice *)READ_BYTES(sizeof(struct wgdevice)))->num_peers;
+	ret = check_version_magic((struct wgdevice *)buffer, ret);
+	if (ret)
+		goto out;
+	for (i = 0; i < len; ++i)
+		READ_BYTES(sizeof(struct wgipmask) * ((struct wgpeer *)READ_BYTES(sizeof(struct wgpeer)))->num_ipmasks);
 	ret = 0;
 out:
-	if (*dev && ret) {
-		free(*dev);
-		*dev = NULL;
+	if (buffer && ret) {
+		free(buffer);
+		buffer = NULL;
 	}
+	*dev = (struct wgdevice *)buffer;
 	close(fd);
 	errno = -ret;
 	return ret;
+
 }
+#undef READ_BYTES
 
 #ifdef __linux__
 static int parse_linkinfo(const struct nlattr *attr, void *data)
@@ -351,7 +374,7 @@ static bool kernel_has_wireguard_interface(const char *interface)
 	char *this_interface;
 	struct inflatable_buffer buffer = { .len = 4096 };
 
-	buffer.buffer = calloc(buffer.len, 1);
+	buffer.buffer = calloc(1, buffer.len);
 	if (!buffer.buffer)
 		return false;
 	if (kernel_get_wireguard_interfaces(&buffer) < 0) {
@@ -372,12 +395,16 @@ static bool kernel_has_wireguard_interface(const char *interface)
 static int do_ioctl(int req, struct ifreq *ifreq)
 {
 	static int fd = -1;
+	int ret;
 	if (fd < 0) {
 		fd = socket(AF_INET, SOCK_DGRAM, 0);
 		if (fd < 0)
 			return fd;
 	}
-	return ioctl(fd, req, ifreq);
+	ret = ioctl(fd, req, ifreq);
+	if (ret == -1)
+		ret = -errno;
+	return ret;
 }
 
 static int kernel_set_device(struct wgdevice *dev)
@@ -385,6 +412,7 @@ static int kernel_set_device(struct wgdevice *dev)
 	struct ifreq ifreq = { .ifr_data = (char *)dev };
 	memcpy(&ifreq.ifr_name, dev->interface, IFNAMSIZ);
 	ifreq.ifr_name[IFNAMSIZ - 1] = 0;
+	dev->version_magic = WG_API_VERSION_MAGIC;
 	return do_ioctl(WG_SET_DEVICE, &ifreq);
 }
 
@@ -400,17 +428,18 @@ static int kernel_get_device(struct wgdevice **dev, const char *interface)
 		ret = do_ioctl(WG_GET_DEVICE, &ifreq);
 		if (ret < 0)
 			goto out;
-		*dev = calloc(ret + sizeof(struct wgdevice), 1);
-		if (!*dev) {
-			ret = -ENOMEM;
+		*dev = calloc(1, ret + sizeof(struct wgdevice));
+		ret = -ENOMEM;
+		if (!*dev)
 			goto out;
-		}
 		(*dev)->peers_size = ret;
+		(*dev)->version_magic = WG_API_VERSION_MAGIC;
 		ifreq.ifr_data = (char *)*dev;
 		memcpy(&ifreq.ifr_name, interface, IFNAMSIZ);
 		ifreq.ifr_name[IFNAMSIZ - 1] = 0;
 		ret = do_ioctl(WG_GET_DEVICE, &ifreq);
 	} while (ret == -EMSGSIZE);
+	ret = check_version_magic(*dev, ret);
 	if (ret < 0) {
 		free(*dev);
 		*dev = NULL;
@@ -428,7 +457,7 @@ char *ipc_list_devices(void)
 	int ret;
 
 	ret = -ENOMEM;
-	buffer.buffer = calloc(buffer.len, 1);
+	buffer.buffer = calloc(1, buffer.len);
 	if (!buffer.buffer)
 		goto cleanup;
 
